import os.path
import sys

simulate_jupyter_content = '''{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0b323301",
   "metadata": {},
   "outputs": [],
   "source": [
    "import ptrade"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9d9779ee",
   "metadata": {},
   "source": [
    "## Rule"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "35a23da8",
   "metadata": {},
   "outputs": [],
   "source": [
    "class MyRule():\\n",
    "    # [交易规则] -----------------------------------------------------------\\n",
    "    # T = 0 -> T0交易\\n",
    "    # T = 1 -> T1交易 ...\\n",
    "    T = 0\\n",
    "    # [时区] --------------------------------------------------------------\\n",
    "    TIMEZONE = 'Asia/Shanghai'  \\n",
    "    # TIMEZONE = 'America/New_York' \\n",
    "    # [账户余额] -----------------------------------------------------------\\n",
    "    ACCOUNT_MONEY = 1000\\n",
    "    # [购买金额] -----------------------------------------------------------\\n",
    "    # 单次开仓的默认金额占比\\n",
    "    BUY_MONEY_RATE = None  \\n",
    "    # 单次开仓的默认金额\\n",
    "    BUY_MONEY = 100  \\n",
    "    # [止盈止损] -----------------------------------------------------------\\n",
    "    # 多单止盈，挂单价格\\n",
    "    LONG_TP_RATE = None  \\n",
    "    # 多单止损，出发止损价格，以市价单平仓\\n",
    "    LONG_SL_RATE = None  \\n",
    "    # 空单止盈，挂单价格\\n",
    "    SHORT_TP_RATE = None  \\n",
    "    # 空单止损，出发止损价格，以市价单平仓\\n",
    "    SHORT_SL_RATE = None  \\n",
    "    # [杠杆] ---------------------------------------------------------------\\n",
    "    # 多单杠杆\\n",
    "    LONG_LEVER = 1  \\n",
    "    # 空单杠杆\\n",
    "    SHORT_LEVER = 1  \\n",
    "    # [仓位] ---------------------------------------------------------------\\n",
    "    # 多单仓位\\n",
    "    LONG_POSITION = 5\\n",
    "    # 空单仓位\\n",
    "    SHORT_POSITION = 5  \\n",
    "    # 总仓位\\n",
    "    POSITION = 10  \\n",
    "    # [最长订单时间（分钟）] ---------------------------------------------------\\n",
    "    # 可以执行开仓的时间\\n",
    "    BUY_PERIODS = [\\n",
    "        ['00:00:00', '23:59:59']  # 可以开仓时间段为00:00:00 ~ 23:59:59\\n",
    "    ]\\n",
    "    # 可以执行平仓的时间\\n",
    "    SELL_PERIODS = [\\n",
    "        ['00:00:00', '23:59:59']  # 可以平仓时间段为00:00:00 ~ 23:59:59\\n",
    "    ]\\n",
    "    # 运行抛出的时间\\n",
    "    THROW_PERIODS = []\\n",
    "    # [最长订单时间（分钟）] ---------------------------------------------------\\n",
    "    MAX_HOLD_ORDER_MINUTE = None\\n",
    "    # [手续费率] -------------------------------------------------------------\\n",
    "    # 开仓的手续费\\n",
    "    BUY_COMMISSION_RATE = 0.00035  \\n",
    "    # 平仓的手续费\\n",
    "    SELL_COMMISSION_RATE = 0.00035  "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9d1e9934",
   "metadata": {},
   "source": [
    "##  Simulate"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "49d99812",
   "metadata": {},
   "outputs": [],
   "source": [
    "class MySimulate(ptrade.Simulate):\\n",
    "    \\n",
    "    def setup(self):\\n",
    "        # do something\\n",
    "        pass\\n",
    "\\n",
    "    def buy(self):\\n",
    "        # do something\\n",
    "        pass\\n",
    "    \\n",
    "    def sell(self):\\n",
    "        # do something\\n",
    "        pass"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f51d677e",
   "metadata": {},
   "source": [
    "## Load Candle"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7e20f53e",
   "metadata": {},
   "outputs": [],
   "source": [
    "candle = ptrade.load.load_candle(\\n",
    "    company = '【交易所】',\\n",
    "    instType = '【产品类别】',\\n",
    "    instId = '【产品ID】',\\n",
    "    start = '【起始日期】',\\n",
    "    end = '【终止日期】',\\n",
    "    bar = '【时间粒度】',\\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "fcd5b23e",
   "metadata": {},
   "source": [
    "## Run "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fd402462",
   "metadata": {},
   "outputs": [],
   "source": [
    "ms = MySimulate(\\n",
    "    candle = candle,\\n",
    "    rule = Rule,\\n",
    "    instId = None,\\n",
    "    warn = True,\\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aecd38a3",
   "metadata": {},
   "source": [
    "## Analysis"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ca41a7dc",
   "metadata": {},
   "outputs": [],
   "source": [
    "# 完整历史交易数据\\n",
    "df = ms.analysis.get_df()\\n",
    "df.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b6882cd1",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_by_position = ms.analysis.get_df_by_position(\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_by_position.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9180a9eb",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_sequential_investment = ms.analysis.get_df_sequential_investment(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    invest_percentage=0.1,  # 投资金额比\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_sequential_investment.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0c3c16d5",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_info_day = ms.analysis.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='day',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_info_day.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "8789944e",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_info_week = ms.analysis.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='week',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_info_week.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "834fc512",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_info_month= ms.analysis.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='month',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_info_month.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "862d1eae",
   "metadata": {},
   "outputs": [],
   "source": [
    "df_info_all= ms.analysis.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='all',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    ")\\n",
    "df_info_all.head()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}'''

project_analysis_jupyter_content = '''{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import pandas as pd\\n",
    "from ptrade.component.analysis import Analysis"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 读取交易数据"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "filepath = './simulate.csv' # 数据路径\\n",
    "df = pd.read_csv(filepath)\\n",
    "df.head(2)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## 分析"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "ans = Analysis()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 按照仓位计算交易过程"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_by_position = ans.get_df_by_position(\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_by_position.head()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 按照等比投资计算交易过程"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_sequential_investment = ans.get_df_sequential_investment(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    invest_percentage=0.1,  # 投资金额比\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_sequential_investment.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "### 按阶段分析交易过程数据"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_info_day = ans.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='day',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_info_day.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_info_week = ans.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='week',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_info_week.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_info_month= ans.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='month',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_info_month.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": [
    "df_info_all= ans.get_df_process_info(\\n",
    "    account_money=1000,  # 初始账户金额\\n",
    "    period='all',  # 统计的单位时间段\\n",
    "    position=None,  # 最大总仓位\\n",
    "    longPosition=None,  # 最大多仓位\\n",
    "    shortPosition=None,  # 最大空仓位\\n",
    "    df=df,  # 数据\\n",
    ")\\n",
    "df_info_all.head()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "pycharm": {
     "name": "#%%\\n"
    }
   },
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 1
}'''

project_rule_content = '''import ptrade.component.rule


class MyRule(ptrade.component.rule.Rule):
    # [交易规则] -----------------------------------------------------------
    # T = 0 -> T0交易
    # T = 1 -> T1交易 ...
    T = 0
    # [时区] --------------------------------------------------------------
    TIMEZONE = 'Asia/Shanghai'
    # TIMEZONE = 'America/New_York'
    # [账户余额] -----------------------------------------------------------
    ACCOUNT_MONEY = 1000
    # [购买金额] -----------------------------------------------------------
    # 单次开仓的默认金额占比
    BUY_MONEY_RATE = None
    # 单次开仓的默认金额
    BUY_MONEY = 100
    # [止盈止损] -----------------------------------------------------------
    # 多单止盈，挂单价格
    LONG_TP_RATE = None
    # 多单止损，出发止损价格，以市价单平仓
    LONG_SL_RATE = None
    # 空单止盈，挂单价格
    SHORT_TP_RATE = None
    # 空单止损，出发止损价格，以市价单平仓
    SHORT_SL_RATE = None
    # [杠杆] ---------------------------------------------------------------
    # 多单杠杆
    LONG_LEVER = 1
    # 空单杠杆
    SHORT_LEVER = 1
    # [仓位] ---------------------------------------------------------------
    # 多单仓位
    LONG_POSITION = 5
    # 空单仓位
    SHORT_POSITION = 5
    # 总仓位
    POSITION = 10
    # [最长订单时间（分钟）] ---------------------------------------------------
    # 可以执行开仓的时间
    BUY_PERIODS = [
        ['00:00:00', '23:59:59']  # 可以开仓时间段为00:00:00 ~ 23:59:59
    ]
    # 可以执行平仓的时间
    SELL_PERIODS = [
        ['00:00:00', '23:59:59']  # 可以平仓时间段为00:00:00 ~ 23:59:59
    ]
    # 运行抛出的时间
    THROW_PERIODS = []
    # [最长订单时间（分钟）] ---------------------------------------------------
    MAX_HOLD_ORDER_MINUTE = None
    # [手续费率] -------------------------------------------------------------
    # 开仓的手续费
    BUY_COMMISSION_RATE = 0.00035
    # 平仓的手续费
    SELL_COMMISSION_RATE = 0.00035'''

project_simulate_content = '''import ptrade
from rule import MyRule


class MySimulate(ptrade.Simulate):
    def __init__(
            self,
            candle,
            rule,
            instId: str = None,
            warn: bool = True,
    ):
        super(MySimulate, self).__init__(candle, rule, instId, warn)

    def setup(self):
        # do something
        pass

    def buy(self):
        # do something
        pass

    def sell(self):
        # do something
        pass


if __name__ == '__main__':
    # ------------------------------------------------------
    instId = '【产品ID】'
    load_candle_param = dict(
        company='【交易所】',
        instType='【产品类别】',
        instId=instId,
        start='【起始日期】',
        end='【终止日期】',
        bar='【时间粒度】',
    )
    filepath = './simulate.csv'  # 交易数据存储路径
    # ------------------------------------------------------
    candle = ptrade.load.load_candle(**load_candle_param)
    ms = MySimulate(
        candle=candle,
        rule=MyRule,
        instId=instId,
        warn=True
    )
    ms.run()
    df = ms.analysis.get_df()
    df.to_csv(filepath)
    print(df.head())'''

project_simulate_multiply_content = '''import ptrade
from rule import MyRule
from simulate import MySimulate



if __name__ == '__main__':
    # ------------------------------------------------------
    instIds = [] # 产品ID列表
    load_candle_map_param = dict(
        company='【交易所】',
        instType='【产品类别】',
        instIds=instIds,
        start='【起始日期】',
        end='【终止日期】',
        bar='【时间粒度】',
    )

    filepath = './simulate_multiply.csv'  # 交易数据存储路径
    # ------------------------------------------------------
    candle_map = ptrade.load.load_candle_map(**load_candle_map_param)
    sm = ptrade.SimulateMultiply(
        Simulate=MySimulate,
        candle_map=candle_map,
        rule = MyRule,
        warn= True
    )
    sm.run(p_num=4)
    df = sm.analysis.get_df()
    df.to_csv(filepath)
    print(df.head())'''


def execute():
    commands = sys.argv
    if len(commands) == 2 and commands[1].lower() == '--help':
        msg = 'ptrade jupyter <name>    创建jupyter shortcuts\nptrade startproject <name>    创建 ptrade项目'
        print(msg)
    elif commands[1].lower() == 'jupyter':
        filename = commands[2]
        if not filename.endswith('.ipynb'):
            filename = filename + '.ipynb'
        with open(filename, 'w') as f:
            f.write(simulate_jupyter_content)
    elif commands[1].lower() == 'startproject':
        dirname = commands[2]
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(os.path.join(dirname, 'analysis.ipynb'), 'w') as f:
            f.write(project_analysis_jupyter_content)
        with open(os.path.join(dirname, 'rule.py'), 'w') as f:
            f.write(project_rule_content)
        with open(os.path.join(dirname, 'simulate.py'), 'w') as f:
            f.write(project_simulate_content)
        with open(os.path.join(dirname, 'simulate.py'), 'w') as f:
            f.write(project_simulate_multiply_content)
        


from ptrade.simulate.simulate import Simulate
from ptrade.simulate.simulate_multiply import SimulateMultiply


# from typing import Union
# import datetime
# import numpy as np
# import paux.process
# from ptrade.simulate._trade import _Trade as __Trade
# import ptrade.component
#
#
# class Simulate(__Trade):
#     def __init__(
#             self,
#             candle: np.array,
#             rule: object,
#             instId: str = None,
#             warn: bool = True,
#     ):
#         '''
#         :param candle: 历史K线数据
#         :param rule: 规则
#         :param instId: 产品ID
#         :param warn: 是否报告警告
#         '''
#         # 参数与内置模块
#         self.candle = candle  # K线数据
#         self.rule = rule  # 规则
#         self.instId = instId  # 标底
#         self.warn = warn
#         self._order = ptrade.component.Order(accountMoney=rule.ACCOUNT_MONEY)  # 订单模块
#         # 当前K线单位时间内
#         self.index = np.nan  # 被遍历的candle行索引
#         self.open = np.nan  # 开盘价
#         self.low = np.nan  # 最低价
#         self.high = np.nan  # 最高价
#         self.close = np.nan  # 收盘价
#         self.volume = np.nan  # todo
#         self.ts = np.nan  # 开盘时刻的毫秒级时间戳
#         # 当前时刻
#         self.datetime = None
#         self.date = None
#         self.time = None
#         # 前一天的数据
#         self.last_day_open = np.nan  # 昨日开盘价
#         self.last_day_high = np.nan  # 昨日最高价
#         self.last_day_low = np.nan  # 昨日最低价
#         self.last_day_close = np.nan  # 昨日收盘价
#
#         self.setup()
#
#
# def _process_simualte_multiply_worker(
#         simulate,start,end,clear
# ):
#     simulate.run(start=start,end=end,clear=clear)
#     return [simulate.historyOrderDatas,simulate.currentOrderDatas]
#
#
#
# # 执行多个产品的模拟交易
# class SimulateMultiply():
#     def __init__(
#             self,
#             Simulate,
#             candle_map: dict,
#             rule: object,
#             warn: bool = True,
#     ):
#         self.simulate_map = {}
#         for instId, candle in candle_map.items():
#             self.simulate_map[instId] = Simulate(
#                 candle=candle,
#                 rule=rule,
#                 instId=instId,
#                 warn=warn
#             )
#
#     def run(
#             self,
#             start: Union[datetime.date, int, float, str, None] = None,
#             end: Union[datetime.date, int, float, str, None] = None,
#             clear: bool = True,
#             p_num=1,
#             skip_exception=False,
#     ):
#         if p_num <= 1:
#             self.historyOrderDatas = []
#             for instId, simulate in self.simulate_map.items():
#                 simulate.run(start=start, end=end, clear=clear)
#                 self.historyOrderDatas += simulate.historyOrderDatas
#             self.analysis = ptrade.component.Analysis(historyOrderDatas=self.historyOrderDatas)
#         else:
#             # warnings.warn('p_num应该为1，异步执行存在安全性问题，下个版本会更新')
#             params = []
#             index_to_instId = {}
#             for index, (instId, simulate) in enumerate(self.simulate_map.items()):
#                 index_to_instId[index] = instId
#                 params.append(
#                     {
#                         'simulate':simulate,
#                         'start': start,
#                         'end': end,
#                         'clear': clear,
#                     }
#                 )
#             results = paux.process.pool_worker(
#                 params=params,
#                 p_num=p_num,
#                 func=_process_simualte_multiply_worker,
#                 skip_exception=skip_exception
#             )
#             self.historyOrderDatas = []
#             self.currentOrderDatas = []
#             for index, (historyOrderDatas,currentOrderDatas) in enumerate(results):
#                 instId = index_to_instId[index]
#                 self.simulate_map[instId].historyOrderDatas = historyOrderDatas
#                 self.simulate_map[instId].currentOrderDatas = currentOrderDatas
#                 self.historyOrderDatas += historyOrderDatas
#                 self.historyOrderDatas += currentOrderDatas
#             self.analysis = ptrade.component.Analysis(historyOrderDatas=self.historyOrderDatas)



class __BaseError(Exception):
    name = ''

    def __init__(self, msg, *args):
        self.msg = '【%s】' % self.name
        self.msg += msg
        for arg in args:
            self.msg += str(arg)

    def __str__(self):
        return self.msg


class PriceError(__BaseError):
    name = 'PriceError'


class OrderError(__BaseError):
    name = 'orderError'


class ParamError(__BaseError):
    name = 'ParamError'


class MoneyError(__BaseError):
    name = 'MoneyError'


class PosSideParamError(ParamError):
    def __init__(self, posSide):
        msg = f'posSide={posSide}，posSide应该属于["long","short"]'
        super(PosSideParamError, self).__init__(msg=msg)


class ToModeParamError(ParamError):
    def __init__(self, toMode):
        msg = f'toMode={toMode}，toMode应该属于["cross","isolated"]'
        super(ToModeParamError, self).__init__(msg=msg)


class TraderError(__BaseError):
    name = 'TraderError'


class UnexpectError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


from typing import Union
import datetime
import paux.process
import ptrade.component


def _process_simualte_multiply_worker(
        simulate, start, end, clear
):
    simulate.run(start=start, end=end, clear=clear)
    return [simulate.historyOrderDatas, simulate.currentOrderDatas]


# 执行多个产品的模拟交易
class SimulateMultiply():
    def __init__(
            self,
            Simulate,
            candle_map: dict,
            rule: object,
            warn: bool = True,
    ):
        self.simulate_map = {}
        for instId, candle in candle_map.items():
            self.simulate_map[instId] = Simulate(
                candle=candle,
                rule=rule,
                instId=instId,
                warn=warn
            )

    def run(
            self,
            start: Union[datetime.date, int, float, str, None] = None,
            end: Union[datetime.date, int, float, str, None] = None,
            clear: bool = True,
            p_num=1,
            skip_exception=False,
    ):
        if p_num <= 1:
            self.historyOrderDatas = []
            for instId, simulate in self.simulate_map.items():
                simulate.run(start=start, end=end, clear=clear)
                self.historyOrderDatas += simulate.historyOrderDatas
            self.analysis = ptrade.component.Analysis(historyOrderDatas=self.historyOrderDatas)
        else:
            # warnings.warn('p_num应该为1，异步执行存在安全性问题，下个版本会更新')
            params = []
            index_to_instId = {}
            for index, (instId, simulate) in enumerate(self.simulate_map.items()):
                index_to_instId[index] = instId
                params.append(
                    {
                        'simulate': simulate,
                        'start': start,
                        'end': end,
                        'clear': clear,
                    }
                )
            results = paux.process.pool_worker(
                params=params,
                p_num=p_num,
                func=_process_simualte_multiply_worker,
                skip_exception=skip_exception
            )
            self.historyOrderDatas = []
            self.currentOrderDatas = []
            for index, (historyOrderDatas, currentOrderDatas) in enumerate(results):
                instId = index_to_instId[index]
                self.simulate_map[instId].historyOrderDatas = historyOrderDatas
                self.simulate_map[instId].currentOrderDatas = currentOrderDatas
                self.historyOrderDatas += historyOrderDatas
                self.historyOrderDatas += currentOrderDatas
            self.analysis = ptrade.component.Analysis(historyOrderDatas=self.historyOrderDatas)



import numpy as np
import ptrade.component
from ptrade.simulate._trade import _Trade as __Trade


class Simulate(__Trade):
    def __init__(
            self,
            candle: np.array,
            rule: object,
            instId: str = None,
            warn: bool = True,
    ):
        '''
        :param candle: 历史K线数据
        :param rule: 规则
        :param instId: 产品ID
        :param warn: 是否报告警告
        '''
        # 参数与内置模块
        self.candle = candle  # K线数据
        self.rule = rule  # 规则
        self.instId = instId  # 标底
        self.warn = warn
        self._order = ptrade.component.Order(accountMoney=rule.ACCOUNT_MONEY)  # 订单模块
        # 当前K线单位时间内
        self.index = np.nan  # 被遍历的candle行索引
        self.open = np.nan  # 开盘价
        self.low = np.nan  # 最低价
        self.high = np.nan  # 最高价
        self.close = np.nan  # 收盘价
        self.volume = np.nan  # 交易量
        self.ts = np.nan  # 开盘时刻的毫秒级时间戳
        # 当前时刻
        self.datetime = None
        self.date = None
        self.time = None
        # 前一天的数据
        self.last_day_open = np.nan  # 昨日开盘价
        self.last_day_high = np.nan  # 昨日最高价
        self.last_day_low = np.nan  # 昨日最低价
        self.last_day_close = np.nan  # 昨日收盘价

        self.setup()



'''
记录数据：
属性                 类型         说明
index               int         被遍历的candle行索引
open                float       开盘价（以时间粒度为单位，如bar='1m'，open为每分钟的开盘价格）
low                 float       最低价
high                float       最高价
close               float       收盘价
volume              float       交易量
ts                  float       开盘时刻的毫秒时间戳

datetime            datetime    当前日期时间
date                date        当前日期
time                time        当前时间

this_day_open       float       今日开盘价
this_day_open_ts    float       今日开盘毫秒时间戳

last_day_open       float       昨日开盘价
last_day_high       float       昨日最高价
last_day_low        float       昨日最低价
last_day_close      float       昨日收盘价
last_day_volume     float       昨日成交量
last_day_open_ts    float       昨日开盘毫秒时间戳

last_day_candle     np.array    昨日K线数据
'''

import datetime
import numpy as np
import paux.order
import paux.date
import paux.candle.transform
from ptrade.simulate._property import _Property


class _Record(_Property):
    # 当前K线单位时间内
    index: int  # 被遍历的candle行索引
    open = float  # 开盘价（以时间粒度为单位，如bar='1m'，open为每分钟的开盘价格）
    low = float  # 最低价
    high = float  # 最高价
    close = float  # 收盘价
    volume = float  # 交易量
    ts: float  # 开盘时刻的毫秒时间戳
    # 当前时刻
    datetime: datetime.datetime  # 当前日期时间
    date: datetime.date  # 当前日期
    time: datetime.time  # 当前时间
    # 当天数据
    this_day_open: float  # 今日开盘价
    this_day_open_ts: float  # 今日开盘毫秒时间戳
    # 前一天的数据
    last_day_open: float  # 昨日开盘价
    last_day_high: float  # 昨日最高价
    last_day_low: float  # 昨日最低价
    last_day_close: float  # 昨日收盘价
    last_day_volume: float  # 昨日成交额
    last_day_open_ts: float  # 昨日开盘毫秒时间戳

    # 记录数据
    def record(self):
        '''
        更新:
        1. 每个时刻的价格数据
        2. 每日开盘价
        3. 昨日的开盘、最高、最低与收盘价
        '''
        # 当前颗粒度中的数据
        self.ts = self.candle[self.index, 0]  # 开盘时刻毫秒级时间戳
        self.open = self.candle[self.index, 1]  # 开盘价
        self.high = self.candle[self.index, 2]  # 最高价
        self.low = self.candle[self.index, 3]  # 最低价
        self.close = self.candle[self.index, 4]  # 收盘价
        self.volume = self.candle[self.index, 5]  # 交易量
        # 当前日期时间
        self.datetime = paux.date.to_datetime(
            date=self.ts,
            timezone=self.rule.TIMEZONE,
        )
        # 当前日期
        self.date = self.datetime.date()
        # 当前时间
        self.time = self.datetime.time()

        if not hasattr(self, 'last_date'):
            self.__last_date = self.date
            self.last_date = None
            self.last_day_open = np.nan  # 昨日开盘价
            self.last_day_high = np.nan  # 昨日最高价
            self.last_day_low = np.nan  # 昨日最低价
            self.last_day_close = np.nan  # 昨日收盘价
            self.last_day_volume = np.nan  # 昨日成交额

        if self.date != self.__last_date:
            self.last_date = self.__last_date
            self.__last_date = self.date
            self.this_day_open = self.open  # 今日开盘价
            self.this_day_open_ts = self.ts  # 今日开盘毫秒时间戳
            # 昨日开盘毫秒时间戳
            self.last_day_open_ts = paux.date.to_ts(
                date=self.last_date,
                timezone=self.rule.TIMEZONE,
            )
            # 昨日历史K线
            self.last_day_candle = self.candle[
                (self.candle[:, 0] >= self.last_day_open_ts) & (self.candle[:, 0] < self.this_day_open_ts)
                ]
            self.last_day_open = self.last_day_candle[0, 1]  # 昨日开盘价
            self.last_day_high = self.last_day_candle[:, 2].max()  # 昨日最高价
            self.last_day_low = self.last_day_candle[:, 3].min()  # 昨日最低价
            self.last_day_close = self.last_day_candle[:, 4]  # 昨日收盘价
            self.last_day_volume = self.last_day_candle[:, 5].sum()  # 昨日成交额



'''
订单中包含的数据
start_orderData带有的参数：

orderId             str             订单ID
orderType           str             订单类型
instId              str             产品ID
toMode              str             全仓|逐仓   全仓：cross 逐仓：isolated
posSide             str             持仓方向    做多：long 做空：short
lever               int             杠杆 现货交易相当于杠杆为1
orderDatetime       str             开仓日期时间  格式：%Y-%m-%d %H:%M:%S
buyLine             int|float       开仓价格
buyCommissionRate   int|float       购买时候的手续费率
tpLine              int|float|None  止盈价格
slLine              int|float|None  止损价格
smLine              int|float|None  爆仓价格
buyMoney            int|float       开仓花费的金额

close_orderData计算的参数：

endType             str                 平仓方式    止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
endDatetime         str                 平仓日期时间  格式：%Y-%m-%d %H:%M:%S
sellLine            int|float           平仓价格
sellCommissionRate  int|float           卖出时候的手续费率
sellMoney           int|float           平仓后得到的金额（以考虑手续费）
holdMinute          int                 持仓时间（分钟）
commission          int|float           手续费
profit              int|float           利润
profitRate          int|float           利润率
hh                  int|float           持仓过程中的最高价格
hhRate              int|float           持仓过程中的最高价格相对于开仓价格的上涨百分比
ll                  int|float           持仓过程中的最低价格
llRate              int|float           持仓过程中的最低价格相对于开仓价格的下跌百分比

辅助字段：
orderIndex          int                 开仓Candle索引
endIndex            int                 平仓Candle索引
orderTs             float               开仓毫秒时间戳
endTs               float               平仓毫秒时间戳
'''
from typing import Literal, Union
import datetime
import paux.order
from ptrade import exception


class Order():
    def __init__(self, accountMoney: Union[int, float]):
        # 持仓总订单
        self.currentOrderDatas = []
        # 持仓多单
        self.currentLongOrderDatas = []
        # 持仓空单
        self.currentShortOrderDatas = []
        # 历史订单
        self.historyOrderDatas = []
        # 账户余额
        self.accountMoney = accountMoney

    @staticmethod
    def get_empty_orderData():
        orderData = {
            # start_orderData带有的参数：
            # orderId             str             订单ID
            'orderId': None,
            # orderType           str             订单类型
            'orderType': None,
            # instId              str             产品ID
            'instId': None,
            # toMode              str             全仓|逐仓   全仓：cross 逐仓：isolated
            'toMode': None,
            # posSide             str             持仓方向    做多：long 做空：short
            'posSide': None,
            # lever               int             杠杆 现货交易相当于杠杆为1
            'lever': None,
            # orderDatetime       str             开仓日期时间  格式：%Y-%m-%d %H:%M:%S
            'orderDatetime': None,
            # buyLine             int|float       开仓价格
            'buyLine': None,
            # buyCommissionRate   int|float       购买时候的手续费率
            'buyCommissionRate': None,
            # tpLine              int|float|None  止盈价格
            'tpLine': None,
            # slLine              int|float|None  止损价格
            'slLine': None,
            # smLine              int|float|None  爆仓价格
            'smLine': None,
            # buyMoney            int|float       开仓花费的金额
            'buyMoney': None,

            # close_orderData计算的参数：
            # endType             str                 平仓方式    止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
            'endType': None,
            # endDatetime         str                 平仓日期时间  格式：%Y-%m-%d %H:%M:%S
            'endDatetime': None,
            # sellLine            int|float           平仓价格
            'sellLine': None,
            # sellCommissionRate  int|float           卖出时候的手续费率
            'sellCommissionRate': None,
            # sellMoney           int|float           平仓后得到的金额（以考虑手续费）
            'sellMoney': None,
            # holdMinute          int                 持仓时间（分钟）
            'holdMinute': None,
            # commission          int|float           手续费
            'commission': None,
            # profit              int|float           利润
            'profit': None,
            # profitRate          int|float           利润率
            'profitRate': None,
            # hh                  int|float           持仓过程中的最高价格
            'hh': None,
            # hhRate              int|float           持仓过程中的最高价格相对于开仓价格的上涨百分比
            'hhRate': None,
            # ll                  int | float           持仓过程中的最低价格
            'll': None,
            # llRate              int|float           持仓过程中的最低价格相对于开仓价格的下跌百分比
            'llRate': None,

            # 辅助字段：
            # orderIndex          int                 开仓Candle索引
            'orderIndex': None,
            # endIndex            int                 平仓Candle索引
            'endIndex': None,
            # orderTs             float               开仓毫秒时间戳
            'orderTs': None,
            # endTs               float               平仓毫秒时间戳
            'endTs': None,
        }
        return orderData

    # 开仓
    def start_orderData(
            self,
            orderId: str,
            instId: str,
            toMode: Literal['cross', 'isolated'],
            posSide: Literal['long', 'short'],
            lever: int,
            orderDatetime: object,
            buyLine: Union[int, float],
            buyCommissionRate: Union[int, float],
            buyMoney: Union[int, float],
            tpRate: Union[int, float, None] = None,
            tpLine: Union[int, float, None] = None,
            slRate: Union[int, float, None] = None,
            slLine: Union[int, float, None] = None,
            smLine: Union[int, float, None] = None,
            orderType: object = None,
            **kwargs
    ):
        '''
        开仓
        关键字               类型            说明
        orderId             str             订单ID
        instId              str             产品ID
        toMode              str             全仓|逐仓   全仓：cross 逐仓：isolated
        posSide             str             持仓方向    做多：long 做空：short
        lever               int             杠杆 现货交易相当于杠杆为1
        orderDatetime       str             开仓日期时间  格式：%Y-%m-%d %H:%M:%S
        buyLine             int|float       开仓价格
        buyCommissionRate   int|float       购买时候的手续费率
        buyMoney            int|float       开仓花费的金额
        tpRate              int|float|None  止盈率
        tpLine              int|float|None  止盈价格（优先级：止盈价格 >>> 止盈率）
        slRate              int|float       止损率
        slLine              int|float|None  止损价格（优先级：止损价格 >>> 止损率）
        smLine              int|float|None  爆仓价格（逐仓会自动计算出强制平仓价格，全仓需要用户手动设置smLine）
        orderType           str             订单类型
        '''
        # 获得一个初始化的订单
        orderData = self.get_empty_orderData()
        # 如果没有tpLine，则使用tpRate计算tpLine
        if not tpLine and tpRate:
            if posSide == 'long':
                tpLine = buyLine * (1 + abs(tpRate))
            elif posSide == 'short':
                tpLine = buyLine * (1 - abs(tpRate))
            else:
                raise exception.PosSideParamError(posSide)

        # 如果没有slLine，则使用slRate计算slLine
        if not slLine and slRate:
            if posSide == 'long':
                slLine = buyLine * (1 - abs(slRate))
            elif posSide == 'short':
                slLine = buyLine * (1 + abs(slRate))
            else:
                raise exception.PosSideParamError(posSide)
        # 计算smLine
        if toMode == 'isolated':
            if posSide == 'long':
                smLine = buyLine * (1 - 1 / lever)  # stock market
            elif posSide == 'short':
                smLine = buyLine * (1 + 1 / lever)  # stock market
            else:
                raise exception.PosSideParamError(posSide)
        elif toMode == 'cross':
            smLine = smLine
        else:
            raise exception.ToModeParamError(toMode)
        # 赋值
        orderData['orderId'] = orderId
        orderData['orderType'] = orderType
        orderData['instId'] = instId
        orderData['toMode'] = toMode
        orderData['posSide'] = posSide
        orderData['lever'] = lever
        orderData['orderDatetime'] = orderDatetime
        orderData['buyLine'] = paux.order.round_simulate(buyLine)
        orderData['buyCommissionRate'] = buyCommissionRate
        # 保留位数
        orderData['tpLine'] = paux.order.round_simulate(tpLine) if tpLine != None else None
        orderData['slLine'] = paux.order.round_simulate(slLine) if slLine != None else None
        orderData['smLine'] = paux.order.round_simulate(smLine) if smLine != None else None
        orderData['buyMoney'] = buyMoney
        # 补充字段
        for key, value in kwargs.items():
            orderData[key] = value
        return orderData

    # 平仓
    def close_orderData(
            self,
            orderData: dict,
            sellLine: Union[int, float],
            sellCommissionRate: Union[int, float],
            endDatetime: str,
            endType: str = None,
            **kwargs
    ):
        '''
        平仓
        参数                 类型                 说明
        orderData           dict                待平仓订单
        sellLine            int|float           平仓价格
        sellCommissionRate  int|float           卖出时候的手续费率
        endDatetime         str                 平仓日期时间  格式：%Y-%m-%d %H:%M:%S
        endType             str                 平仓方式    止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
        '''
        # 卖出价格
        sellLine = paux.order.round_simulate(sellLine)
        # 是否需要考虑爆仓
        if orderData['toMode'] == 'cross':
            pass
        elif orderData['toMode'] == 'isolated':
            # 多仓爆仓
            if orderData['posSide'] == 'long' and sellLine <= orderData['smLine']:
                sellLine = orderData['smLine']
                endType = 'SM'
            # 空仓爆仓
            elif orderData['posSide'] == 'short' and sellLine >= orderData['smLine']:
                sellLine = orderData['smLine']
                endType = 'SM'
        else:
            raise exception.ToModeParamError(orderData['toMode'])
        # 持仓时间（分钟）
        holdMinute = int(
            (datetime.datetime.strptime(endDatetime, '%Y-%m-%d %H:%M:%S', ).timestamp() - datetime.datetime.strptime(
                orderData['orderDatetime'], '%Y-%m-%d %H:%M:%S').timestamp()) / 60)
        commission_data = paux.order.get_commission_data(
            posSide=orderData['posSide'],
            buyMoney=orderData['buyMoney'],
            buyLine=orderData['buyLine'],
            sellLine=sellLine,
            lever=orderData['lever'],
            buyCommissionRate=orderData['buyCommissionRate'],
            sellCommissionRate=sellCommissionRate
        )

        commission = commission_data['commission']
        sellMoney = commission_data['sellMoney']
        profitRate = commission_data['profitRate']
        profit = round(profitRate * orderData['buyMoney'], 4)

        # 赋值
        orderData['endType'] = endType
        orderData['endDatetime'] = endDatetime
        orderData['sellLine'] = sellLine
        orderData['holdMinute'] = holdMinute
        orderData['commission'] = commission
        orderData['sellMoney'] = sellMoney
        orderData['profitRate'] = profitRate
        orderData['profit'] = profit
        orderData['sellCommissionRate'] = sellCommissionRate
        # 补充字段
        for key, value in kwargs.items():
            orderData[key] = value
        return orderData

    # 处理平仓数据
    def to_historyOrderDatas(self, orderDatas: Union[list, tuple, None]):
        '''
        按照orderId，删除currentOrderDatas、currentLongOrderDatas、currentShortOrderDatas
        添加到historyOrderDatas中
        账户金额如果出现负数，会报告异常

        orderDatas          list        待平仓订单列表
        '''

        if orderDatas:
            for orderData in orderDatas:
                # 总持仓中的orderId
                currentOrderIds = [orderData['orderId'] for orderData in self.currentOrderDatas]
                # 待平仓的订单Id
                orderId = orderData['orderId']
                # 此订单不在持仓中
                if not orderId in currentOrderIds:
                    continue

                # 删除总持仓中的orderId订单
                del self.currentOrderDatas[currentOrderIds.index(orderId)]
                # 删除多仓持仓中的orderId订单
                if orderData['posSide'] == 'long':
                    # 多仓持仓中的orderId
                    currentLongOrderIds = [orderData['orderId'] for orderData in self.currentLongOrderDatas]
                    del self.currentLongOrderDatas[currentLongOrderIds.index(orderId)]
                # 删除空仓持仓中的orderId订单
                elif orderData['posSide'] == 'short':
                    # 空仓持仓中的orderId
                    currentShortOrderIds = [orderData['orderId'] for orderData in self.currentShortOrderDatas]
                    del self.currentShortOrderDatas[currentShortOrderIds.index(orderId)]
                # 添加到历史订单中
                self.historyOrderDatas.append(orderData)
                # 维护账户余额
                self.accountMoney += orderData['sellMoney']
                if self.accountMoney < 0:
                    raise exception.OrderError('账户余额不能小于零')

    # 进入持仓数据
    def to_currentOrderDatas(self, orderDatas: Union[list, tuple, None]):
        '''
        添加到currentOrderDatas、currentLongOrderDatas、currentShortOrderDatas
        账户金额如果出现负数，会报告异常

        orderDatas          list        开仓订单列表
        '''
        if orderDatas:
            for orderData in orderDatas:
                # 添加持仓
                self.currentOrderDatas.append(orderData)
                if orderData['posSide'] == 'long':
                    self.currentLongOrderDatas.append(orderData)
                elif orderData['posSide'] == 'short':
                    self.currentShortOrderDatas.append(orderData)
                # 维护账户余额
                self.accountMoney -= orderData['buyMoney']
                if self.accountMoney < 0:
                    raise exception.OrderError('账户余额不能小于零')



'''
交易过程：

tpSell_and_slThrow      止盈止损卖出（优先止损）
smThrow                 爆仓
periodThrow             在抛出时间段卖出
timeoutThrow            超过最大持仓时间卖出
marketThrow             以市场价卖出

run                     运行计算
'''

from typing import Union
import datetime
import warnings
import paux.order
import paux.date
import paux.candle.transform
import paux.order
from ptrade import component
from ptrade import exception
from ptrade.simulate._shortcuts import _ShortCuts


class _Trade(_ShortCuts):
    # 仓位不足，是否发送警告
    warn: bool

    PRIORITY = {
        'periodThrow': 250,  # 在抛出时间段平仓
        'timeoutThrow': 200,  # 超过最大持仓时间平仓
        'smThrow': 150,  # 爆仓
        'slThrow': 125,  # 止损平仓
        # 'sell': 100,  # 用户自定义卖出
        'tpSell': 25,  # 止盈平仓
        'marketThrow': 0,  # 以收盘市场价平仓
        # 'buy': 50,  # 用户自定义买入
    }

    # 多种平仓方式---------------------------------------------------------------------------------------------
    # 止盈止损平仓
    def tpSell_and_slThrow(self):
        '''
        止盈或止损平仓
            悲观策略：如果单位时间内价格剧烈波动，同时达到止盈与止损条件，以止损价格抛出
            止盈价格与止损价格并不一定是实际的卖出价格
            止盈卖出：endType = 'TP'
            止损抛出：endType = 'SL'
        '''
        # 过滤非平仓时间
        if not paux.date.is_period_allowed(
                date=self.time,
                periods=self.rule.SELL_PERIODS,
                timezone=self.rule.TIMEZONE,
        ):
            return None
        # 止盈卖出与止损抛出
        for orderData in self.currentOrderDatasCanBeTraded:
            posSide = orderData['posSide']
            # 如果达到了止损条件
            # 多单：最小价<=止损价格；空单：最大价格>=止损价格
            if orderData['slLine'] and (
                    (posSide == 'long' and self.low <= orderData['slLine']) or
                    (posSide == 'short' and self.high >= orderData['slLine'])):
                # 多单实际止损价格
                if posSide == 'long':
                    realSlLine = min(self.open, orderData['slLine'])
                # 空单实际止损价格
                elif posSide == 'short':
                    realSlLine = max(self.open, orderData['slLine'])
                else:
                    raise exception.PosSideParamError(posSide)
                priority = self.PRIORITY['slThrow']  # 止损平仓优先级，默认125
                # 平仓
                orderData = self.close_orderData(
                    orderData=orderData,
                    sellLine=realSlLine,
                    endType='SL',
                    priority=priority,
                )
                yield orderData
                continue
            # 未达成止损条件，观测是否达到止盈条件
            # 多单：最高价>=止盈价格；空单：最低价<=止盈价格
            if orderData['tpLine'] and (
                    (posSide == 'long' and self.high >= orderData['tpLine']) or
                    (posSide == 'short' and self.low <= orderData['tpLine'])):
                # 多单实际止盈价格
                if posSide == 'long':
                    realTpLine = max(self.open, orderData['tpLine'])
                # 空单实际止盈价格
                elif posSide == 'short':
                    realTpLine = min(self.open, orderData['tpLine'])
                else:
                    raise exception.PosSideParamError(posSide)
                # 平仓
                priority = self.PRIORITY['tpSell']  # 止盈平仓优先级，默认50
                orderData = self.close_orderData(
                    orderData=orderData,
                    sellLine=realTpLine,
                    endType='TP',
                    priority=priority,
                )
                yield orderData
                continue

    # 爆仓
    def smThrow(self):
        '''
        逐仓会自动计算出强制平仓价格
        全仓需要用户手动设置smLine
        endType = 'SM'
        '''
        priority = self.PRIORITY['smThrow']  # 爆仓优先级，默认150
        for orderData in self.currentOrderDatasCanBeTraded:
            if (orderData['posSide'] == 'long' and self.low <= orderData['smLine'] or
                    orderData['posSide'] == 'short' and self.high >= orderData['smLine']):
                yield self.close_orderData(
                    orderData=orderData,
                    sellLine=orderData['smLine'],
                    endType='SM',
                    priority=priority,
                )

    # 在抛出时间段平仓
    def periodThrow(self):
        '''
        时间段满足要求，则全部平仓
        endType = 'PT'
        '''
        priority = self.PRIORITY['periodThrow']  # 在抛出时间段平仓，默认250
        # 是否满足时间段要求
        if not paux.date.is_period_allowed(
                date=self.time,
                periods=self.rule.THROW_PERIODS,
                timezone=self.rule.TIMEZONE,
        ):
            return None
        # 全部持仓以市价平掉
        for orderData in self.currentOrderDatasCanBeTraded:
            orderData = self.close_orderData(
                orderData=orderData,
                sellLine=self.open,
                endType='PT',
                priority=priority,
            )
            yield orderData

    # 超过最大持仓时间平仓
    def timeoutThrow(self):
        '''
        持仓时间超时，以市价平仓
        endType = 'TT'
        '''
        priority = self.PRIORITY['timeoutThrow']  # 超过最大持仓时间平仓，默认200
        # 时间段是否允许卖出
        if not paux.date.is_period_allowed(
                date=self.time,
                periods=self.rule.SELL_PERIODS,
                timezone=self.rule.TIMEZONE,
        ):
            return None
        # 无最大持仓时间限制
        if not self.rule.MAX_HOLD_ORDER_MINUTE:
            return None
        # 遍历全部持仓订单
        for orderData in self.currentOrderDatasCanBeTraded:
            # 持仓时间超过了预设最大时间
            if ((self.ts - orderData['orderTs']) / 60000) >= self.rule.MAX_HOLD_ORDER_MINUTE:
                orderData = self.close_orderData(
                    orderData=orderData,
                    sellLine=self.open,
                    endType='TT',
                    priority=priority,
                )
                yield orderData

    # 以收盘市场价平仓
    def marketThrow(self, orderDatas):
        '''
        :param orderDatas:市价平仓的订单
        endType = 'MT'
        '''
        priority = self.PRIORITY['marketThrow']  # 以收盘市场价平仓，默认0
        if orderDatas:
            for orderData in orderDatas:
                orderData = self.close_orderData(
                    orderData=orderData,
                    sellLine=self.close,
                    endType='MT',
                    priority=priority,
                )
                yield orderData

    # 运行计算---------------------------------------------------------------------------------------------
    def run(
            self,
            start: Union[datetime.date, int, float, str, None] = None,
            end: Union[datetime.date, int, float, str, None] = None,
            clear: bool = True,
    ):
        '''
        :param start: 执行模拟交易的起始日期时间
        :param end: 执行模拟交易的终止日期时间
        :param clear: 是否在遍历后的最后时刻市价平掉全部仓位
        优先级：periodThrow > timeoutThrow > smThrow >  slThrow > tpSell > sell > buy
        '''

        # 起始索引与终止索引
        start_index = paux.candle.transform.get_candle_index_by_date(
            candle=self.candle,
            date=start,
            default=0
        )
        # 终止索引
        end_index = paux.candle.transform.get_candle_index_by_date(
            candle=self.candle,
            date=end,
            default=self.candle.shape[0]
        )
        for index in range(start_index, end_index):
            # 更新数据
            self.index = index
            self.record()
            # 时间段抛出
            periodThrow = self.periodThrow()
            periodThrowDatas = [orderData for orderData in periodThrow if periodThrow]
            # 超时抛出
            timeoutThrow = self.timeoutThrow()
            timeoutThrowDatas = [orderData for orderData in timeoutThrow if timeoutThrow]
            # 爆仓强平
            smThrow = self.smThrow()
            smThrowDatas = [orderData for orderData in smThrow if smThrow]
            # 止盈卖出与止损抛出
            tpSell_and_slThrow = self.tpSell_and_slThrow()
            tpSell_and_slThrowDatas = [orderData for orderData in tpSell_and_slThrow if tpSell_and_slThrow]

            # 自定义卖出
            if paux.date.is_period_allowed(
                    date=self.time,
                    periods=self.rule.SELL_PERIODS,
                    timezone=self.rule.TIMEZONE,
            ):
                sell = self.sell()
                if sell:
                    sellDatas = [orderData for orderData in sell if sell]
                else:
                    sellDatas = []
            else:
                sellDatas = []
            # 自定义买入
            if paux.date.is_period_allowed(
                    date=self.time,
                    periods=self.rule.BUY_PERIODS,
                    timezone=self.rule.TIMEZONE,
            ) and not paux.date.is_period_allowed(
                date=self.time,
                periods=self.rule.THROW_PERIODS,
                timezone=self.rule.TIMEZONE
            ):
                buy = self.buy()
                if buy:
                    buyDatas = [orderData for orderData in buy]
                else:
                    buyDatas = []
            else:
                buyDatas = []
            # 订单排序
            totalOrderDatas = periodThrowDatas + timeoutThrowDatas + smThrowDatas + tpSell_and_slThrowDatas + sellDatas + buyDatas
            totalOrderDatas = sorted(totalOrderDatas, key=lambda d: d['priority'], reverse=True)

            for orderData in totalOrderDatas:
                # 购买订单
                if orderData['sellLine'] == None:
                    self.__to_currentOrderDatas(orderData)
                else:
                    self._order.to_historyOrderDatas(orderDatas=[orderData, ])
        # 未成交订单强平
        if clear:
            marketThrow = self.marketThrow(orderDatas=self._order.currentOrderDatas)
            if marketThrow:
                marketThrowOrderDatas = [orderData for orderData in marketThrow]
            else:
                marketThrowOrderDatas = []

            self._order.to_historyOrderDatas(
                marketThrowOrderDatas
            )
        self.analysis = component.Analysis(historyOrderDatas=self.historyOrderDatas)

    # 根据条件判断是否满足进入持仓
    def __to_currentOrderDatas(self, orderData):

        if self.longPositionLeft == 0 and self.shortPositionLeft == 0:
            return None
        if self.longPositionLeft < 0 or self.shortPositionLeft < 0:
            msg = f'longPositionLeft = {self.longPositionLeft},shortPositionLeft={self.shortPositionLeft}'
            raise exception.TraderError(msg=msg)
        if orderData['posSide'] == 'long':
            if self.longPositionLeft > 0:
                self._order.to_currentOrderDatas(orderDatas=[orderData, ])
            else:
                if self.warn:
                    msg = '多仓仓位不足，orderData无法开仓,orderData={orderData}'.format(orderData=str(orderData))
                    warnings.warn(msg)
        elif orderData['posSide'] == 'short':
            if self.shortPositionLeft > 0:
                self._order.to_currentOrderDatas(orderDatas=[orderData, ])
            else:
                if self.warn:
                    msg = '空仓仓位不足，orderData无法开仓,orderData={orderData}'.format(orderData=str(orderData))
                    warnings.warn(msg)



'''
模拟交易中的属性:

属性                                 返回类型     说明
accountMoney                        int|float   账户可用余额
positionAccountMoney                int|float   持仓余额估计
longPositionAccountMoney            int|float   多单持仓余额估计
shortPositionAccountMoney           int|float   空单持仓余额估计

currentOrderDatas                   list        持仓订单数据
currentLongOrderDatas               list        持仓多单数据
currentShortOrderDatas              list        持仓空单数据

currentOrderDatasCanBeTraded        list        可以被交易的订单
currentLongOrderDatasCanBeTraded    list        可以被交易的多单
currentShortOrderDatasCanBeTraded   list        可以被交易的空单

currentTodayOrderDatas              list        今天的仓位数据
currentTodayLongOrderDatas          list        今天的多仓位数据
currentTodayShortOrderDatas         list        今天的空仓位数据

historyOrderDatas                   list        历史订单数据
historyLongOrderDatas               list        历史多单数据
historyShortOrderDatas              list        历史空单数据

currentPosition                     int         获得持仓的总订单数量
currentLongPosition                 int         获得持仓的多单数量
currentShortPosition                int         获得持仓的空单数量

longPositionLeft                    int         剩余多仓数量
shortPositionLeft                   int         剩余空仓数量
'''

import numpy as np
import datetime
from ptrade.component.order import Order as _Order
from ptrade.component.rule import Rule
import paux.order


class _Property():
    candle: np.array
    instId: str
    rule: Rule
    _order: _Order
    date: datetime.date

    # 余额 ------------------------------------------------------------------------------------------------
    # 账户可用余额
    @property
    def accountMoney(self):
        return self._order.accountMoney

    @accountMoney.setter
    def accountMoney(self, value):
        self._order.accountMoney = value

    # 持仓余额估计
    def __positionAccountMoney(self, orderDatas):
        accountMoney_in_position = 0
        for orderData in orderDatas:
            sellMoney = paux.order.get_commission_data(
                posSide=orderData['posSide'],
                buyMoney=orderData['buyMoney'],
                buyLine=orderData['buyLine'],
                sellLine=orderData['sellLine'],
                lever=orderData['lever'],
                buyCommissionRate=self.rule.BUY_COMMISSION_RATE,
                sellCommissionRate=self.rule.SELL_COMMISSION_RATE,
            )['sellMoney']
            accountMoney_in_position += sellMoney
        return accountMoney_in_position

    # 持仓余额估计
    @property
    def positionAccountMoney(self):
        orderDatas = self.currentOrderDatas
        return self.__positionAccountMoney(orderDatas)

    # 多单持仓余额估计
    @property
    def longPositionAccountMoney(self):
        orderDatas = self.currentLongOrderDatas
        return self.__positionAccountMoney(orderDatas)

    # 空单持仓余额估计
    @property
    def shortPositionAccountMoney(self):
        orderDatas = self.currentShortOrderDatas
        return self.__positionAccountMoney(orderDatas)

    # 持仓订单数据---------------------------------------------------------------------------------------------
    # 持仓订单数据
    @property
    def currentOrderDatas(self):
        return self._order.currentOrderDatas

    @currentOrderDatas.setter
    def currentOrderDatas(self, value):
        self._order.currentOrderDatas = value

    # 持仓多单数据
    @property
    def currentLongOrderDatas(self):
        return self._order.currentLongOrderDatas

    # 持仓空单数据
    @property
    def currentShortOrderDatas(self):
        return self._order.currentShortOrderDatas

    # 可以被交易的订单
    def __currentOrderDatasCanBeTraded(self, orderDatas):
        if self.rule.T == 0:
            return orderDatas
        dateCanBeTrade = (self.date + datetime.timedelta(days=self.rule.T)).strftime('%Y-%m-%d')

        orderDatas = []
        for orderData in orderDatas:
            if dateCanBeTrade >= orderData['orderDatetime']:
                orderDatas.append(orderData)
        return orderDatas

    # 可以被交易的订单
    @property
    def currentOrderDatasCanBeTraded(self):
        orderDatas = self.currentOrderDatas
        return self.__currentOrderDatasCanBeTraded(orderDatas)

    # 可以被交易的多单
    @property
    def currentLongOrderDatasCanBeTraded(self):
        orderDatas = self.currentLongOrderDatas
        return self.__currentOrderDatasCanBeTraded(orderDatas)

    # 以被交易的空单
    @property
    def currentShortOrderDatasCanBeTraded(self):
        orderDatas = self.currentShortOrderDatas
        return self.__currentOrderDatasCanBeTraded(orderDatas)

    # 今天的仓位
    def __currentTodayOrderDatas(self, orderDatas):
        return [
            orderData for orderData in orderDatas if
            orderData['orderDatetime'][0:10] == self.date.strftime('%Y-%m-%d')
        ]

    # 今天的仓位数据
    @property
    def currentTodayOrderDatas(self):
        orderDatas = self.currentOrderDatas
        return self.__currentTodayOrderDatas(orderDatas)

    # 今天的多仓位数据
    @property
    def currentTodayLongOrderDatas(self):
        orderDatas = self.currentLongOrderDatas
        return self.__currentTodayOrderDatas(orderDatas)

    # 今天的空仓位数据
    @property
    def currentTodayShortOrderDatas(self):
        orderDatas = self.currentShortOrderDatas
        return self.__currentTodayOrderDatas(orderDatas)

    # 历史订单数据---------------------------------------------------------------------------------------------
    # 历史订单数据
    @property
    def historyOrderDatas(self):
        return self._order.historyOrderDatas

    @historyOrderDatas.setter
    def historyOrderDatas(self, value):
        self._order.historyOrderDatas = value

    # 历史多单数据
    @property
    def historyLongOrderDatas(self):
        return [
            orderData for orderData in self._order.historyOrderDatas if orderData['posSide'] == 'long'
        ]

    # 历史空单数据
    @property
    def historyShortOrderDatas(self):
        return [
            orderData for orderData in self._order.historyOrderDatas if orderData['posSide'] == 'short'
        ]

    # 仓位数量---------------------------------------------------------------------------------------------
    # 获得持仓的总订单数量
    @property
    def currentPosition(self):
        return len(self._order.currentOrderDatas)

    # 获得持仓的多单数量
    @property
    def currentLongPosition(self):
        return len(self._order.currentLongOrderDatas)

    # 获得持仓的空单数量
    @property
    def currentShortPosition(self):
        return len(self._order.currentShortOrderDatas)

    # 剩余多仓数量
    @property
    def longPositionLeft(self):
        longPositionLeft = min(
            self.rule.LONG_POSITION - self.currentLongPosition,
            self.rule.POSITION - self.currentPosition,
        )
        return longPositionLeft

    # 剩余空仓数量
    @property
    def shortPositionLeft(self):
        # 剩余仓位数量
        shortPositionLeft = min(
            self.rule.SHORT_POSITION - self.currentShortPosition,
            self.rule.POSITION - self.currentPosition,
        )
        return shortPositionLeft



'''
1. 可以重写的函数：

1.1 init后执行的方法
setup()

1.2 买入策略
buy()
    return None # 结束函数
    yield self.start_orderData(...) # 开仓

1.3 卖出策略
sell()
    return None # 结束函数
    yield self.close_orderData(...) # 平仓


2. 可以调用的函数
2.1 开仓
start_orderData(
    self,
    buyLine: Union[int, float],
    posSide: Literal['long', 'short'],
    buyMoney: Union[int, float,Literal['rule']] = 'rule',
    buyCommissionRate: Union[int, float, Literal['rule']] = 'rule',
    lever: Union[int, Literal['rule']] = 'rule',
    tpRate: Union[int, float, Literal['rule']] = 'rule',
    tpLine: Union[int, float, None] = None,
    slRate: Union[int, float, Literal['rule']] = 'rule',
    slLine: Union[int, float, None] = None,
    orderType: str = None,
    toMode: Literal['isolated', 'cross'] = 'isolated',
    priority=100,
    **kwargs
)

参数                 类型             说明
buyLine             int|float       开仓价格
posSide             str             持仓方向，多仓：long 空仓：short
buyMoney            int|float|str   开仓金额，默认值：rule
buyCommissionRate   int|float|str   开仓手续费率，默认值：rule
lever               int|str         杠杆
tpRate              int|float|str   止盈率，默认值：rule
tpLine              int|float|None  止盈价格，默认值为None，tpLine与tpRate同时存在，优先级：tpLine >>> tpRate
slRate              int|float|str   止损率，默认值：rule
slLine              int|float|None  止损价格，默认值为None，slLine与slRate同时存在，优先级：slLine >>> slRate
smLine              int|float|None  爆仓价格，默认值为None，逐仓的爆仓价格会自动算出，无需填写，全仓的爆仓价格需要用户指定，否则不会考虑全仓爆仓
orderType           str             订单类型，默认值：''
toMode              str             持仓类型，全仓：cross 逐仓：isolated
priority            int             订单优先级（用于订单任务调度排序，默认值：100）

2.2 平仓
close_orderData(
    self,
    orderData: dict,
    sellLine: Union[int, float],
    sellCommissionRate: Union[int, float, Literal['rule']] = 'rule',
    endType: object = None,
    priority=50,
    **kwargs
)
参数                     类型             说明
orderData               dict            待平仓的持仓订单
sellLine                int|float       平仓价格
sellCommissionRate      int|float|str   平仓手续费率，默认值：rule
endType                 str             平仓类型，用户调用平仓：CS 止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
priority                int             订单优先级（用于订单任务调度排序，默认值：50）
'''

from typing import Union, Literal
from copy import deepcopy
import uuid
import numpy as np
import paux.order
import paux.date
import paux.candle.transform
import paux.order
from ptrade import exception
from ptrade.simulate._record import _Record


class _ShortCuts(_Record):

    # 用户重写 --------------------------------------------------------------------------------------------
    # init后执行的方法
    def setup(self):
        # to overwrite
        pass

    # 买入策略
    def buy(self):
        # to overwrite
        pass

    # 卖出策略
    def sell(self):
        pass

    # 获取近期指定数量的历史K线数据
    def get_latest_candle(self, length: int, include_now: bool = False):
        '''
        :param length: 长度
        :param include_now: 是否包含当前时刻
        '''
        start_index = self.index - length
        end_index = self.index
        if include_now:
            start_index += 1
            end_index += 1
        return self.candle[start_index:end_index]

    # 获取当天的历史K线数据
    def get_today_candle(self, include_now: bool = False):
        '''
        :param include_now:  是否包含当前时刻
        '''
        this_day_open_ts = paux.date.to_ts(date=self.date, timezone=self.rule.TIMEZONE)
        end_ts = self.datetime.timestamp() * 1000
        if not include_now:
            end_ts -= 1
        return self.candle[
            (self.candle[:, 0] >= this_day_open_ts) & (self.candle[:, 0] <= end_ts)
            ]

    # 开仓订单---------------------------------------------------------------------------------------------
    def start_orderData(
            self,
            buyLine: Union[int, float],
            posSide: Literal['long', 'short'],
            buyMoney: Union[int, float, Literal['rule']] = 'rule',
            buyCommissionRate: Union[int, float, Literal['rule']] = 'rule',
            lever: Union[int, Literal['rule']] = 'rule',
            tpRate: Union[int, float, Literal['rule']] = 'rule',
            tpLine: Union[int, float, None] = None,
            slRate: Union[int, float, Literal['rule']] = 'rule',
            slLine: Union[int, float, None] = None,
            orderType: str = '',
            toMode: Literal['isolated', 'cross'] = 'isolated',
            priority=50,
            **kwargs
    ):
        '''
        发起订单
        字段                 类型             说明
        buyLine             int|float       开仓价格
        posSide             str             持仓方向，多仓：long 空仓：short
        buyMoney            int|float|str   开仓金额，默认值：rule
        buyCommissionRate   int|float|str   开仓手续费率，默认值：rule
        lever               int|str         杠杆
        tpRate              int|float|str   止盈率，默认值：rule
        tpLine              int|float|None  止盈价格，默认值为None，tpLine与tpRate同时存在，优先级：tpLine >>> tpRate
        slRate              int|float|str   止损率，默认值：rule
        slLine              int|float|None  止损价格，默认值为None，slLine与slRate同时存在，优先级：slLine >>> slRate
        smLine              int|float|None  爆仓价格，默认值为None，逐仓的爆仓价格会自动算出，无需填写，全仓的爆仓价格需要用户指定，否则不会考虑全仓爆仓
        orderType           str             订单类型，默认值：''
        toMode              str             持仓类型，全仓：cross 逐仓：isolated
        priority            int             订单优先级（用于订单任务调度排序，默认值：100）

        随机生成32位长度的orderId
        '''
        orderId = uuid.uuid4().hex[0:32]
        # 杠杆
        if lever == 'rule':
            if posSide == 'long':
                lever = self.rule.LONG_LEVER
            elif posSide == 'short':
                lever = self.rule.SHORT_LEVER
            else:
                raise exception.PosSideParamError(posSide)
        # 止盈率
        if tpRate == 'rule':
            if posSide == 'long':
                tpRate = self.rule.LONG_TP_RATE
            elif posSide == 'short':
                tpRate = self.rule.SHORT_TP_RATE
        # 止损率
        if slRate == 'rule':
            if posSide == 'long':
                slRate = self.rule.LONG_SL_RATE
            elif posSide == 'short':
                slRate = self.rule.SHORT_SL_RATE
        # 当前订单的candle行索引
        orderIndex = self.index
        # 开仓价格，修正价格属于最小值和最大值之间
        buyLine = paux.order.round_simulate(np.clip(buyLine, self.low, self.high))
        # 开仓手续费率
        if buyCommissionRate == 'rule':
            buyCommissionRate = self.rule.BUY_COMMISSION_RATE
        # 如果没有buyMoney
        if buyMoney == 'rule':
            if self.rule.BUY_MONEY_RATE:
                buyMoney = np.clip(
                    (self.positionAccountMoney + self.accountMoney) * self.rule.BUY_COMMISSION_RATE,
                    0, self.accountMoney
                )
            else:
                buyMoney = self.rule.BUY_MONEY

        return self._order.start_orderData(
            orderId=orderId,  # 订单ID
            orderType=orderType,  # 订单类型
            instId=self.instId,  # 产品ID
            lever=lever,  # 杠杆
            toMode=toMode,  # 持仓类型
            posSide=posSide,  # 持仓方向
            buyMoney=buyMoney,  # 开仓金额
            orderDatetime=self.datetime.strftime('%Y-%m-%d %H:%M:%S'),  # 开仓日期时间
            buyLine=buyLine,  # 开仓价格
            buyCommissionRate=buyCommissionRate,  # 开仓手续费率
            tpLine=tpLine,  # 止盈率
            tpRate=tpRate,  # 止盈价格
            slLine=slLine,  # 止损率
            slRate=slRate,  # 止损价格
            smLine=None,  # 爆仓价格
            orderTs=self.ts,  # 补充字段
            orderIndex=orderIndex,  # 补充字段
            priority=priority,  # 优先级
            **kwargs,
        )

    # 平仓订单---------------------------------------------------------------------------------------------
    def close_orderData(
            self,
            orderData: dict,
            sellLine: Union[int, float],
            sellCommissionRate: Union[int, float, Literal['rule']] = 'rule',
            endType: str = 'CS',
            priority=100,
            **kwargs
    ):
        '''
        结束订单
        orderData               dict            待平仓的持仓订单
        sellLine                int|float       平仓价格
        sellCommissionRate      int|float|str   平仓手续费率，默认值：rule
        endType                 str             平仓类型，用户调用平仓：CS 止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
        priority                int             订单优先级（用于订单任务调度排序，默认值：50）

        辅助字段：
            endIndex        int         平仓时刻的candle索引
            endTs           float       平仓时刻的毫秒时间戳
        通过计算产生的字段：
            commission      int|float   手续费
            holdMinute      int         持仓时间（分钟）
            sellMoney       int|float   平仓后得到的金额（以考虑手续费）
            profitRate      int|float   利润率
            profit          int|float   利润
        补充的过程数据：
            hh              int|float   持仓过程中的最高价格
            hhRate          int|float   持仓过程中的最高价格相对于开仓价格的上涨百分比
            hh              int|float   持仓过程中的最低价格
            hhRate          int|float   持仓过程中的最低价格相对于开仓价格的下跌百分比
        '''
        orderData = deepcopy(orderData)
        # 手续费
        if sellCommissionRate == 'rule':
            sellCommissionRate = self.rule.SELL_COMMISSION_RATE
        # 开始与终止索引
        endIndex = self.index
        orderIndex = orderData['orderIndex']
        # 过程中的最高价与最低价
        if orderIndex != endIndex:
            hh = self.candle[orderIndex:endIndex, 2].max()
            ll = self.candle[orderIndex:endIndex, 3].min()
        else:
            hh = self.candle[endIndex, 2]
            ll = self.candle[endIndex, 3]
        # 最高价与最低价相对于buyLine的增长下跌比例
        hhRate = round((hh - orderData['buyLine']) / orderData['buyLine'], 4)
        llRate = round((ll - orderData['buyLine']) / orderData['buyLine'], 4)
        # 平仓
        return self._order.close_orderData(
            orderData=orderData,  # 待平仓的持仓订单
            sellLine=sellLine,  # 平仓价格
            endDatetime=self.datetime.strftime('%Y-%m-%d %H:%M:%S'),  # 平仓日期时间
            sellCommissionRate=sellCommissionRate,  # 平仓手续费率
            endType=endType,  # 平仓类型
            # 补充的字段
            endIndex=endIndex,  # 平仓时刻的candle索引
            endTs=self.ts,  ## 平仓时刻的毫秒时间戳
            hh=hh,  # 持仓过程中的最高价格
            hhRate=hhRate,  # 持仓过程中的最高价格相对于开仓价格的上涨百分比
            ll=ll,  # 持仓过程中的最低价格
            llRate=llRate,  # 持仓过程中的最低价格相对于开仓价格的下跌百分比
            priority=priority,  # 订单优先级
            **kwargs
        )



'''
get_df      获取历史订单：


orderType       str         订单类型
posSide         str         持仓方向    做多：long 做空：short
instId          str         产品ID
toMode          str         全仓|逐仓   全仓：cross 逐仓：isolated
lever           int         杠杆 现货交易相当于杠杆为1
orderDatetime   str         开仓日期时间  格式：%Y-%m-%d %H:%M:%S
endDatetime     str         平仓日期时间  格式：%Y-%m-%d %H:%M:%S
holdMinute      int         持仓时间（分钟）
endType         str         平仓方式    止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT
buyLine         int|float   开仓价格
sellLine        int|float   平仓价格
tpLine          int|float|None   止盈价格
slLine          int|float|None   止损价格
smLine          int|float|None   爆仓价格
buyMoney        int|float   开仓花费的金额
sellMoney       int|float   平仓后得到的金额（以考虑手续费）
commission      int|float   手续费
profit          int|float   利润
profitRate      int|float   利润率
hh              int|float   持仓过程中的最高价格
hhRate          int|float   持仓过程中的最高价格相对于开仓价格的上涨百分比
ll              int|float           持仓过程中的最低价格
llRate          int|float           持仓过程中的最低价格相对于开仓价格的下跌百分比


INFO_COLUMNS
period                  str         时间段
profit_sum              int|float   总利润
profit_mean             int|float   平均单位订单利润
profit_peak             int|float   利润最高峰
profit_valley           int|float   利润最低谷
dprofit                 int|float   最大连续盈利
dloss                   int|float   最大连续亏损
profit_num              int|float   盈利订单数量
loss_num                int|float   亏损订单数量
profit_suc_rate         int|float   盈利订单占比
order_num               int         订单数量
account_money           int|float   时间段结尾账户金额
account_money_peak      int|float   账户金额最高峰
account_money_valley    int|float   账户金额最低谷
'''

'''
|orderType|str|订单类型|
|posSide|str|持仓方向 做多：long 做空：short|
|instId|str|产品ID|
|toMode|str|全仓|逐仓 全仓：cross 逐仓：isolated|
|lever|int|杠杆，现货交易相当于杠杆为1|
|orderDatetime|str|开仓日期时间 格式：%Y-%m-%d %H:%M:%S|
|endDatetime|str|平仓日期时间 格式：%Y-%m-%d %H:%M:%S|
|holdMinute|int|持仓时间（分钟）|
|endType|str|平仓方式，止盈：TP 止损：SL 爆仓：SM 按照时间段卖出：PT 超过最大持仓时间：TT 模拟交易后以市场价平仓：MT|
|buyLine|int \| float|开仓价格|
|sellLine|int \| float|平仓价格|
|tpLine|int \| float \| None|止盈价格|
|slLine|int \| float \| None|止损价格|
|smLine|int \| float \| None|爆仓价格|
|buyMoney|int \| float|开仓花费的金额|
|sellMoney|int \| float|平仓后得到的金额（以考虑手续费）|
|commission|int \| float|手续费|
|profit|int \| float|利润|
|profitRate|int \| float|利润率|
|hh|int \| float|持仓过程中的最高价格|
|hhRate|int \| float|持仓过程中的最高价格相对于开仓价格的上涨百分比|
|ll|int \| float|持仓过程中的最低价格|
|llRate|int \| float|持仓过程中的最低价格相对于开仓价格的下跌百分比|
'''

from typing import Literal, Union
from copy import deepcopy
import numpy as np
import pandas as pd
import datetime
from ptrade import exception

pd.set_option('display.max_rows', 100)  # 最多展示1000行
pd.set_option('display.max_columns', 25)  # 最多展示10列


# 分析模块
class Analysis():
    # 默认展示的DataFrame字段
    DETAIL_COLUMNS = [
        'orderType',
        'posSide',
        'instId',
        'toMode',
        'lever',
        'orderDatetime',
        'endDatetime',
        'holdMinute',
        'endType',
        'buyLine',
        'sellLine',
        'tpLine',
        'slLine',
        'smLine',
        'buyMoney',
        'sellMoney',
        'commission',
        'profit',
        'profitRate',
        'hh',
        'hhRate',
        'll',
        'llRate',
    ]

    # 过程数据中展示的字段
    INFO_COLUMNS = [
        'period',
        'profit_sum',
        'profit_mean',
        'profit_peak',
        'profit_valley',
        'dprofit',
        'dloss',
        'profit_num',
        'loss_num',
        'profit_suc_rate',
        'order_num',
        'account_money',
        'account_money_peak',
        'account_money_valley',
    ]

    def __init__(self, historyOrderDatas: list = []):
        self.historyOrderDatas = historyOrderDatas
        if self.historyOrderDatas:
            self.df = pd.DataFrame(self.historyOrderDatas).sort_values(by='orderDatetime').reset_index(
                drop=True)  # 保留更多的字段
        else:
            self.df = pd.DataFrame([], columns=self.DETAIL_COLUMNS)



    def __get_df_columns(self, **kwargs):
        # 默认字段
        df_columns = deepcopy(self.DETAIL_COLUMNS)
        # 自定义控制字段的显示
        for column, show in kwargs.items():
            if show == True and column not in df_columns:
                df_columns.append(column)
                continue
            if show == False and column in df_columns:
                df_columns.remove(column)
                continue
        return df_columns

    # 获得历史订单的DataFrame对象
    def get_df(
            self,
            historyOrderDatas: list = [],
            **kwargs
    ):
        '''
        :param historyOrderDatas: 历史订单数据
        :param kwargs: 补充字段的显示与隐藏

        优先级：
        historyOrderDatas >>> self.df
        '''
        if not historyOrderDatas:
            df = self.df
        else:
            df = pd.DataFrame(historyOrderDatas).sort_values(by='orderDatetime').reset_index(drop=True)
        # 筛选字段
        df_columns = self.__get_df_columns(**kwargs)
        df = df[df_columns]
        return df

    # 按照仓位计算交易过程DataFrame
    def get_df_by_position(
            self,
            position: int = None,
            longPosition: int = None,
            shortPosition: int = None,
            historyOrderDatas: list = [],
            df: pd.DataFrame = None,
            **kwargs,
    ):
        '''
        :param position: 总仓位
        :param longPosition: 多仓位
        :param shortPosition: 空仓位
        :param historyOrderDatas: 列表类型的历史订单
        :param df: DataFrame类型的历史订单
        :param kwargs: 补充字段的显示与隐藏

        优先级：
            historyOrderDatas >>> df >>> self.df >>> self.historyOrderDatas
        '''
        # >>1 使用historyOrderDatas
        if historyOrderDatas:
            df = self.get_df(historyOrderDatas=historyOrderDatas)
        # >>2 使用df
        elif type(df).__name__ != 'NoneType':
            df = df
        else:
            # >>3 使用self.df
            if hasattr(self, 'df'):
                df = self.df
            # >> 4 获取self.df
            else:
                df = self.get_df()
        # 空数据
        if df.shape[0] == 0:
            return df
        # 日期排序
        df = df.sort_values(by='orderDatetime').reset_index(drop=True)
        # 如果需要过滤仓位
        if position or longPosition or shortPosition:
            # position longPosition shortPosition默认为无穷大
            if position == None:
                position = 9999 * 9999
            if longPosition == None:
                longPosition = 9999 * 9999
            if shortPosition == None:
                shortPosition = 9999 * 9999
            # 初始化
            currentLongOrderDatas = []
            currentShortOrderDatas = []
            currentOrderDatas = []
            historyOrderDatas = []

            for i in df.index:
                # 本次的订单
                thisOrderData = df.loc[i].to_dict()
                # 卖出之前的订单
                removeLongOrderDatas = []
                removeShortOrderDatas = []
                removeOrderDatas = []
                # 找到应该被平仓的订单
                for orderData in currentOrderDatas:
                    if orderData['endDatetime'] <= thisOrderData['orderDatetime']:
                        removeOrderDatas.append(orderData)
                        if orderData['posSide'] == 'long':
                            removeLongOrderDatas.append(orderData)
                        elif orderData['posSide'] == 'short':
                            removeShortOrderDatas.append(orderData)
                # 平仓 删除与添加(总)
                for orderData in removeOrderDatas:
                    historyOrderDatas.append(orderData)
                    currentOrderDatas.remove(orderData)
                # 平仓 删除(多)
                for orderData in removeLongOrderDatas:
                    currentLongOrderDatas.remove(orderData)
                # 平仓 删除(空)
                for orderData in removeShortOrderDatas:
                    currentShortOrderDatas.remove(orderData)
                # 剩余多仓位
                longPositionLeft = min(
                    position - len(currentOrderDatas),
                    longPosition - len(currentLongOrderDatas),
                )
                # 剩余空仓位
                shortPositionLeft = min(
                    position - len(currentOrderDatas),
                    shortPosition - len(currentShortOrderDatas),
                )
                # 按照是否有剩余仓位，决定是否能添加到current
                if thisOrderData['posSide'] == 'long':
                    if longPositionLeft > 0:
                        currentLongOrderDatas.append(thisOrderData)
                        currentOrderDatas.append(thisOrderData)
                elif thisOrderData['posSide'] == 'short':
                    if shortPositionLeft > 0:
                        currentShortOrderDatas.append(thisOrderData)
                        currentOrderDatas.append(thisOrderData)
            # 结尾强制平仓
            historyOrderDatas += currentOrderDatas
            df = pd.DataFrame(historyOrderDatas)
        # 筛选字段
        df_columns = self.__get_df_columns(**kwargs)
        df = df[df_columns]
        return df

    # 计算等比连续投资的DataFrame
    def get_df_sequential_investment(
            self,
            account_money: Union[int, float],
            invest_percentage: Union[int, float] = 0.1,
            position: int = None,
            longPosition: int = None,
            shortPosition: int = None,
            historyOrderDatas: list = [],
            df: pd.DataFrame = None,
            **kwargs,
    ):
        '''
        :param account_money: 账户余额
        :param invest_percentage: 投资金额比
        :param position: 总仓位
        :param longPosition: 多仓位
        :param shortPosition: 空仓位
        :param historyOrderDatas: 列表类型的历史订单
        :param df: DataFrame类型的历史订单
        :param kwargs: 补充字段的显示与隐藏

        优先级：
            historyOrderDatas >>> df >>> self.df >>> self.historyOrderDatas
        '''
        # 按照仓位过滤得到df
        df = self.get_df_by_position(
            position=position,
            longPosition=longPosition,
            shortPosition=shortPosition,
            historyOrderDatas=historyOrderDatas,
            df=df,
        )
        # 空数据
        if not df.shape[0]:
            return df
        # 日期排序
        df = df.sort_values(by='orderDatetime').reset_index(drop=True)
        # 初始化订单列表
        currentOrderDatas = []
        historyOrderDatas = []
        # 逐行遍历
        for i in df.index:
            # 如果当前持仓超过了总仓位，是未知错误
            if position != None and len(currentOrderDatas) > position:
                error_msg = '未知错误，超过了总仓位'
                raise exception.UnexpectError(error_msg)
            # 本次的订单
            thisOrderData = df.loc[i].to_dict()
            # 可以卖出的订单
            removeOrderDatas = []
            for orderData in currentOrderDatas:
                if orderData['endDatetime'] <= thisOrderData['orderDatetime']:
                    removeOrderDatas.append(orderData)
            # 添加与删除，并添加金额到account_money
            for orderData in removeOrderDatas:
                account_money += orderData['profit']
                historyOrderDatas.append(orderData)
                currentOrderDatas.remove(orderData)
            # 如果有余额
            if account_money > 0:
                # 计算购买余额
                buyMoney = min(
                    account_money,
                    (account_money + sum([
                        orderData['buyMoney'] for orderData in currentOrderDatas
                    ])) * invest_percentage
                )
                # alter比例
                alter_rate = buyMoney / thisOrderData['buyMoney']
                # 等比扩大
                thisOrderData['buyMoney'] = round(alter_rate * thisOrderData['buyMoney'], 4)
                thisOrderData['sellMoney'] = round(alter_rate * thisOrderData['sellMoney'], 4)
                thisOrderData['commission'] = round(alter_rate * thisOrderData['commission'], 8)
                thisOrderData['profit'] = round(alter_rate * thisOrderData['profit'], 4)
                currentOrderDatas.append(thisOrderData)
        # 结尾强制平仓
        historyOrderDatas += currentOrderDatas
        # 有订单数据
        if historyOrderDatas:
            df_sequential_investment = pd.DataFrame(historyOrderDatas)
        # 无订单数据
        else:
            df_sequential_investment = pd.DataFrame([], columns=self.DETAIL_COLUMNS)
        df_columns = self.__get_df_columns(**kwargs)
        df_sequential_investment = df_sequential_investment[df_columns]
        return df_sequential_investment

    # 获取交易过程中的分析数据
    def get_df_process_info(
            self,
            account_money: Union[int, float],
            period: Literal['day', 'week', 'month', 'year', 'all'],
            position: int = None,
            longPosition: int = None,
            shortPosition: int = None,
            historyOrderDatas: list = [],
            df: pd.DataFrame = None,
    ):
        '''
        :param account_money: 账户余额
        :param period: 统计的单位时间段
        :param position: 总仓位
        :param longPosition: 多仓位
        :param shortPosition: 空仓位
        :param historyOrderDatas: 列表类型的历史订单
        :param df: DataFrame类型的历史订单

        优先级：
            historyOrderDatas >>> df >>> self.df >>> self.historyOrderDatas
        '''
        # 按照仓位过滤得到df
        df = self.get_df_by_position(
            position=position,
            longPosition=longPosition,
            shortPosition=shortPosition,
            historyOrderDatas=historyOrderDatas,
            df=df
        )
        # 空数据
        if not df.shape[0]:
            return pd.DataFrame([], columns=self.INFO_COLUMNS)
        # 建立单位时间段字段period
        if period == 'day':
            df[period] = df['endDatetime'].apply(lambda d: d[0:10])
        elif period == 'week':
            df[period] = ''
            for index in df.index:
                this_year, this_month = datetime.datetime.strptime(
                    df.loc[index, 'endDatetime'],
                    '%Y-%m-%d %H:%M:%S'
                ).isocalendar()[0:2]
                df.loc[index, period] = '{:0>4d}~{:0>2d}'.format(this_year, this_month)
        elif period == 'month':
            df[period] = df['endDatetime'].apply(lambda d: d[0:7])
        elif period == 'year':
            df[period] = df['endDatetime'].apply(lambda d: d[0:4])
        elif period == 'all':
            df[period] = '{start_date}~{end_date}'.format(
                start_date=df['orderDatetime'].min()[0:10],
                end_date=df['endDatetime'].min()[0:10],
            )
        else:
            msg = f'period = f{period}，period must in ["day","week","month","year","all"]'
            raise exception.ParamError(msg)
        df['isprofit'] = 0
        df['isloss'] = 0
        for index in df.index:
            if df.loc[index, 'profit'] > 0:
                df.loc[index, 'isprofit'] = 1
            else:
                df.loc[index, 'isloss'] = 1
        # 初始化数据
        info_datas = []
        # 订单终止的类型
        endTypes = df['endType'].unique().tolist()
        for this_period, df_gb in df.groupby(period):
            profit_sum = round(df_gb['profit'].sum(), 4)  # 总利润
            profit_mean = round(df_gb['profit'].mean(), 4)  # 订单平均利润
            dprofit = 0  # 最大连续盈利
            dloss = 0  # 最大连续亏损
            profits = df_gb['profit'].values  # 利润序列
            cum_profits = np.cumsum(profits)  # 利润累加
            profit_peak = cum_profits.max()  # 利润最高峰
            profit_valley = cum_profits.min()  # 利润最低谷
            for i in range(df_gb.shape[0]):
                cum_profits = np.cumsum(profits[i:])  # 分段利润累加
                dprofit = max(dprofit, np.max(cum_profits))  # 最大连续盈利
                dloss = min(dloss, np.min(cum_profits))  # 最大连续亏损

            profit_num = df_gb['isprofit'].sum()
            loss_num = df_gb['isloss'].sum()
            profit_suc_rate = round(profit_num / (profit_num + loss_num), 3)
            order_num = df_gb.shape[0]  # 订单数量
            account_money_peak = account_money + profit_peak  # 账户金额最高峰
            account_money_valley = account_money + profit_valley  # 账户金额最低谷
            account_money = account_money + profit_sum  # 交易后的交易金额

            this_data = {
                'period': period,
                period: this_period,
                'profit_sum': profit_sum,
                'profit_mean': profit_mean,
                'profit_peak': profit_peak,
                'profit_valley': profit_valley,
                'dprofit': dprofit,
                'dloss': dloss,
                'profit_num': profit_num,
                'loss_num': loss_num,
                'profit_suc_rate': profit_suc_rate,
                'order_num': order_num,
                'account_money': account_money,
                'account_money_peak': account_money_peak,
                'account_money_valley': account_money_valley,

            }
            # 补充各个endType的次数
            for endType in endTypes:
                if endType != None:
                    this_data[endType] = df_gb.query('endType==@endType').shape[0]
            info_datas.append(this_data)
        return pd.DataFrame(info_datas)


'''
load_candle_by_date | load_candle     按照日期读取指定产品的历史K线

company         str                             交易所 OKX BINANCE HUOBI IB TDA EAST_MONEY
instType        str                             产品类别
instId          str                             产品ID
start           int|float|str|datetime.date     起始日期时间
end             int|float|str|datetime.date     终止日期时间
base_dir        str|None                        数据存储文件夹，默认为None，使用paux中settings的交易所K线路径
timezone        str|None                        时区，默认为None，使用paux中settings的交易所K线时区
bar             str                             时间粒度
valid_interval  bool                            是否验证K线间隔
valid_start     bool                            是否验证K线时间起点必须等于start
valid_end       bool                            是否验证K线时间终点必须等于end
minus_bar       int                             去掉历史K线结尾N个时间粒度，默认为1，表示不包含end时刻的K线数据

load_candle_map_by_date | load_candle_map       按照日期读取多个产品的历史K线

company         str                             交易所 OKX BINANCE HUOBI IB TDA EAST_MONEY
instType        str                             产品类别
instIds         list                            产品ID，instIds=[]，表示读取产品类别中数据完整的全部产品Id
start           int|float|str|datetime.date     起始日期时间
end             int|float|str|datetime.date     终止日期时间
base_dir        str|None                        数据存储文件夹，默认为None，使用paux中settings的交易所K线路径
timezone        str|None                        时区，默认为None，使用paux中settings的交易所K线时区
bar             str                             时间粒度
valid_interval  bool                            是否验证K线间隔
valid_start     bool                            是否验证K线时间起点必须等于start
valid_end       bool                            是否验证K线时间终点必须等于end
minus_bar       int                             去掉历史K线结尾N个时间粒度，默认为1，表示不包含end时刻的K线数据
'''

import datetime
from typing import Union, Literal

from paux.candle.history import candle_okx
from paux.candle.history import candle_binance
from paux.candle.history import candle_huobi
from paux.candle.history import candle_ib
from paux.candle.history import candle_tda
from paux.candle.history import candle_east_money
import paux.settings

settings = paux.settings.read_settings()

COMPANY_MAP = {
    'OKX': {
        'module': candle_okx,
        'base_dir': settings['CANDLE_OKX_BASE_DIR'],
        'timezone': settings['CANDLE_OKX_TIMEZONE'],
    },
    'BINANCE': {
        'module': candle_binance,
        'base_dir': settings['CANDLE_BINANCE_BASE_DIR'],
        'timezone': settings['CANDLE_BINANCE_TIMEZONE'],
    },
    'HUOBI': {
        'module': candle_huobi,
        'base_dir': settings['CANDLE_HUOBI_BASE_DIR'],
        'timezone': settings['CANDLE_HUOBI_TIMEZONE'],
    },
    'IB': {
        'module': candle_ib,
        'base_dir': settings['CANDLE_IB_BASE_DIR'],
        'timezone': settings['CANDLE_IB_TIMEZONE'],
    },
    'TDA': {
        'module': candle_tda,
        'base_dir': settings['CANDLE_TDA_BASE_DIR'],
        'timezone': settings['CANDLE_TDA_TIMEZONE'],
    },
    'EAST_MONEY': {
        'module': candle_east_money,
        'base_dir': settings['CANDLE_EAST_MONEY_BAST_DIR'],
        'timezone': settings['CANDLE_EAST_MONEY_TIMEZONE'],
    },
}


# 按照日期读取指定产品的历史K线
def load_candle_by_date(
        company: Literal['OKX', 'BINANCE', 'HUOBI', 'IB', 'TDA', 'EAST_MONEY'],
        instType: str,
        instId: str,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: Union[str, None] = None,
        timezone: Union[str, None] = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar: int = 1,
):
    '''
        company         str                             交易所 OKX BINANCE HUOBI IB TDA EAST_MONEY
        instType        str                             产品类别
        instId          str                             产品ID
        start           int|float|str|datetime.date     起始日期时间
        end             int|float|str|datetime.date     终止日期时间
        base_dir        str|None                        数据存储文件夹，默认为None，使用paux中settings的交易所K线路径
        timezone        str|None                        时区，默认为None，使用paux中settings的交易所K线时区
        bar             str                             时间粒度
        valid_interval  bool                            是否验证K线间隔
        valid_start     bool                            是否验证K线时间起点必须等于start
        valid_end       bool                            是否验证K线时间终点必须等于end
        minus_bar       int                             去掉历史K线结尾N个时间粒度，默认为1，表示不包含end时刻的K线数据

        如果instIds = []，读取的产品类型中的全部产品ID后，会进行过滤，仅保留从start~end中均有数据的产品ID
    '''
    module = COMPANY_MAP[company.upper()]['module']
    if base_dir == None:
        base_dir = COMPANY_MAP[company.upper()]['base_dir']
    if timezone == None:
        timezone = COMPANY_MAP[company.upper()]['timezone']
    return module.load_candle_by_date(
        instType=instType,
        instId=instId,
        start=start,
        end=end,
        base_dir=base_dir,
        timezone=timezone,
        bar=bar,
        valid_interval=valid_interval,
        valid_start=valid_start,
        valid_end=valid_end,
        minus_bar=minus_bar,
    )


# 按照日期读取多个产品的历史K线
def load_candle_map_by_date(
        company: Literal['OKX', 'BINANCE', 'HUOBI', 'IB', 'TDA', 'EAST_MONEY'],
        instType: str,
        instIds: list,
        start: Union[int, float, str, datetime.date],
        end: Union[int, float, str, datetime.date],
        base_dir: Union[str, None] = None,
        timezone: Union[str, None] = None,
        bar: Literal['1m', '3m', '5m', '15m', '1H', '2H', '4H'] = '1m',
        p_num=4,
        valid_interval=True,
        valid_start=True,
        valid_end=True,
        minus_bar: int = 1,
):
    '''
        company         str                             交易所 OKX BINANCE HUOBI IB TDA EAST_MONEY
        instType        str                             产品类别
        instIds         list                            产品ID，instIds=[]，表示读取产品类别中数据完整的全部产品Id
        start           int|float|str|datetime.date     起始日期时间
        end             int|float|str|datetime.date     终止日期时间
        base_dir        str|None                        数据存储文件夹，默认为None，使用paux中settings的交易所K线路径
        timezone        str|None                        时区，默认为None，使用paux中settings的交易所K线时区
        bar             str                             时间粒度
        valid_interval  bool                            是否验证K线间隔
        valid_start     bool                            是否验证K线时间起点必须等于start
        valid_end       bool                            是否验证K线时间终点必须等于end
        minus_bar       int                             去掉历史K线结尾N个时间粒度，默认为1，表示不包含end时刻的K线数据
    '''
    module = COMPANY_MAP[company.upper()]['module']
    if base_dir == None:
        base_dir = COMPANY_MAP[company.upper()]['base_dir']
    if timezone == None:
        timezone = COMPANY_MAP[company.upper()]['timezone']
    return module.load_candle_map_by_date(
        instType=instType,
        instIds=instIds,
        start=start,
        end=end,
        base_dir=base_dir,
        timezone=timezone,
        bar=bar,
        p_num=p_num,
        valid_interval=valid_interval,
        valid_start=valid_start,
        valid_end=valid_end,
        minus_bar=minus_bar,
    )


# for shortcuts
load_candle = load_candle_by_date
load_candle_map = load_candle_map_by_date
